#include <stdio.h>
#include <strings.h>
#include <errno.h>
#include <sys/file.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/uio.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <des.h>
#include <krb.h>
#include "network.h"
#include "services.h"

#define MAXCONTEXTS 64 /* use whatever maxfd is; look it up */
#define GROUPS "/site/calserver/groups"
#define NEXT "/site/calserver/messages/next"
#define TRANSPATH "/site/calserver/messages"
#define DISTFULLNAME "/site/calendar/src/server/distribute"
#define DISTNAME "distribute"
#define USERS "/site/calserver/users"

void sendauth();

typedef struct _context
{
  connection *c;

  int authenticated;
  AUTH_DAT ad;
  des_key_schedule sched;

  char *transbuff;        /* pointer to loaded transactions file */
  char *transwhere;       /* where we're reading in the text */

  char *ptr;              /* to hold onto read data */
  int len;

  int srv_err;
  int sub_err;            /* return error if authentication fails */
  char miscbuff[1024];    /* must be per-context */
} context;

static int nextcontext = 0;
static context *contexts[MAXCONTEXTS]; /* inited to zero? */

/*
 * Set nextcontext to a free context number
 */
void bumpcontext()
{
  int start;

  start = nextcontext;

  nextcontext = (nextcontext + 1) % MAXCONTEXTS;
  while (contexts[nextcontext] != NULL && nextcontext != start)
    nextcontext = (nextcontext + 1) % MAXCONTEXTS;

  if (nextcontext == start)
    {
      fprintf(stderr, "toast: context overflow\n");
      exit(1); /* should never happen if MAXCONTEXTS >= maxfd */
    }
}

/*
 * Kill a context and free its context number
 */
killcontext(a)
context *a;
{
  contexts[a->c->context] = NULL;
  i_kill(a->c, -1);
  free(a);
}

/*
 * This routine is called for a new connection on the listening
 * socket. Set up a new context for it.
 */
void newclient(c)
connection *c;
{
  context *a;

  fprintf(stdout, "(%d) New connection on socket %d from host %s\n",
	  nextcontext, c->s, inet_ntoa(c->peer.sin_addr));

  a = contexts[nextcontext] = (context *)malloc(sizeof(context));
  c->context = nextcontext;
  bumpcontext();

  a->c = c;
  a->authenticated = 0;
}

/*
 * Generic memory freeer to be called after write data has
 * been unused
 */
void callfree(c)
connection *c;
{
  free(c->w_buffer);
  c->s_empty = NULL;
}

void dump(ptr, len)
char *ptr;
int len;
{
  int i;

  for (i = 0; i < len; i++)
    putc(ptr[i], stdout);

  putc('\n', stdout);
  fflush(stdout);
}

/*
 * Error returner; called from writeprocs only.
 */
w_error(c, maj, min)
connection *c;
int maj, min;
{
  context *a;

  a = contexts[c->context];

  a->srv_err = htonl(maj);
  a->sub_err = htonl(min);
  c->w_buffer = (char *)&a->srv_err;
  c->w_len = 2*sizeof(a->srv_err);
  gag(c);

  fprintf(stdout, "(%d) Major %d, minor %d\n", c->context, maj, min);
}

/*
 * Error returner; called from writeprocs only.
 */
w_error_continue(c, maj, min)
connection *c;
int maj, min;
{
  context *a;

  a = contexts[c->context];

  a->srv_err = htonl(maj);
  a->sub_err = htonl(min);
  c->w_buffer = (char *)&a->srv_err;
  c->w_len = 2*sizeof(a->srv_err);

  fprintf(stdout, "(%d) Major %d, minor %d\n", c->context, maj, min);
}

/* bug - what if client makes muliple requests without waiting for responses */

/*
 * REQ_GROUPS_ALL
 * Dump the groups file to the client.
 * Let's cache it someday. What a concept!
 * (there's just no incentive to take 5 minutes and do it now)
 */
void reqgroupsall(c)
connection *c;
{
  context *a;
  int fd, len;
  struct stat buf;
  buffer *rtn;

  a = contexts[c->context];

  fd = open(GROUPS, O_RDONLY, 0);

  if (fd == -1)
    {
      w_error(c, ERR_GROUPS_ALL_OPEN, errno);
      return;
    }

  if (-1 == fstat(fd, &buf)) /* could only return EIO */
    {
      w_error(c, ERR_GROUPS_ALL_FSTAT, errno);
      close(fd);
      return;
    }

  len = (int)buf.st_size;

  rtn = (buffer *)malloc(len + sizeof(buffer));
  if (rtn == 0)
    {
      w_error(c, ERR_GROUPS_MALLOC, 0);
      close(fd);
      return;
    } /* let's just ignore the fact that we're possibly in deep shit */

  /* trusting the man page on this one. I know, I know... */
  if (-1 == read(fd, rtn->data, len))
    {
      free(rtn);
      w_error(c, ERR_GROUPS_READ, errno);
      close(fd);
      return;
    }

  close(fd);

  rtn->srv_err = htonl(ERR_GROUPS_ALL_SUCCESS);
  rtn->sub_err = htonl(0);

  c->w_buffer = (char *)rtn;
  c->w_len = len + sizeof(buffer);
  gag(c);
  c->s_write = NULL;
  c->s_empty = callfree;
}

int getnexttrans(name)
char *name;
{
  int f;
  char buffer[20];
  int trans, len;

  f = open(NEXT, O_RDWR, 0);

  if (f < 0)
    return -1;

  if (-1 == (len = read(f, buffer, 19)))
    {
      close(f);
      return -1;
    }

  buffer[len] = '\0';
  trans = atoi(buffer);

  lseek(f, 0, L_SET);
  write(f, buffer, sprintf(buffer, "%d\n", trans + 1));

  /* if this fails, we might consider exiting... */

  close(f);

  sprintf(name, "%s/%d", TRANSPATH, trans);

  return trans;
}

/*
 * Send the actual transactions of REQ_MSG_RECV (follows this routine)
 */
void sendtrans(c)
connection *c;
{
  context *a;
  int fd, len;
  struct stat buf;
  buffer *rtn;
  char f[50];
  int transnum;

  a = contexts[c->context];

  transnum = atoi(a->transwhere);
  a->transwhere = index(a->transwhere, '\n') + 1;
  if (*(a->transwhere) == '\0') /* then we're doing the last one */
    {
      gag(c);
      c->s_write = NULL;
      free(a->transbuff);
    }

  sprintf(f, "%s/%d", TRANSPATH, transnum);
  fd = open(f, O_RDONLY, 0);

  if (fd == -1)
    {
      w_error_continue(c, ERR_MSG_RECV_OPEN, errno);
      return;
    }

  if (-1 == fstat(fd, &buf)) /* could only return EIO */
    {
      w_error_continue(c, ERR_MSG_RECV_FSTAT, errno);
      close(fd);
      return;
    }

  len = (int)buf.st_size;

  rtn = (buffer *)malloc(len + sizeof(buffer));
  if (rtn == 0)
    {
      w_error_continue(c, ERR_MSG_RECV_MALLOC, 0);
      close(fd);
      return;
    } /* let's just ignore the fact that we're possibly in deep shit */

  /* trusting the man page on this one. I know, I know... */
  if (-1 == read(fd, rtn->data, len))
    {
      free(rtn);
      w_error_continue(c, ERR_MSG_RECV_READ, errno);
      close(fd);
      return;
    }

  close(fd);

  rtn->srv_err = htonl(ERR_MSG_RECV_SUCCESS);
  rtn->sub_err = htonl(transnum);

  c->w_buffer = (char *)rtn;
  c->w_len = len + sizeof(buffer);
  c->s_empty = callfree;
}


/*
 * REQ_MSG_RECV
 * Return n+1 messages, where 0 contains <(success,n)>, 1 contains
 * <(success,id1),message1>, 2 contains...
 */
void msgrecv(c)
connection *c;
{
  context *a;
  char f[100];
  int fd;
  struct stat buf;
  int len;
  char *tmp;
  int count = 0;

  a = contexts[c->context];

  /* can't read the messages if they haven't authenticated */
  if (a->authenticated == 0)
    {
      w_error(c, ERR_MSG_RECV_NOTAUTH, errno);
      return;
    }

  sprintf(f, "%s/%s", USERS, a->ad.pname);

  fd = open(f, O_RDONLY, 0);

  if (fd == -1)
    {
      if (errno == ENOENT)
	{
	  w_error(c, ERR_MSG_RECV_SUCCESS, 0); /* no transactions */
	  return;
	}

      w_error(c, ERR_MSG_RECV_OPEN, errno);
      return;
    }

  if (-1 == fstat(fd, &buf))
    {
      w_error(c, ERR_MSG_RECV_FSTAT, errno);
      close(fd);
      return;
    }

  len = (int)buf.st_size;

  a->transbuff = (char *)malloc(len + 1);
  if (a->transbuff == 0)
    {
      w_error(c, ERR_MSG_RECV_MALLOC, 0);
      close(fd);
      return;
    } /* let's just ignore the fact that we're possibly in deep shit */

  /* trusting the man page on this one. I know, I know... */
  if (-1 == read(fd, a->transbuff, len))
    {
      free(a->transbuff);
      w_error(c, ERR_MSG_RECV_READ, errno);
      close(fd);
      return;
    }

  close(fd);

  a->transbuff[len] = '\0'; /* make it a string */
  a->transwhere = a->transbuff;

  tmp = a->transbuff;
  while (*tmp != '\0')
    if (*tmp++ == '\n')
      count++;

  a->srv_err = htonl(ERR_MSG_RECV_SUCCESS);
  a->sub_err = htonl(count);
  c->w_buffer = (char *)&a->srv_err;
  c->w_len = 2*sizeof(a->srv_err);

  if (count)
    c->s_write = sendtrans;
}

/*
 * REQ_MSG_SEND
 * Parse and distribute the message, leave the transaction id# to
 * be returned
 */
void msgsend(c)
connection *c;
{
  context *a;
  int trans, transnum;
  char *ptr;
  char *e;
  char name[100];
  char buffer[100]; /* should NOT be a problem */
  int r, l;
  int ferr;

  a = contexts[c->context];

  if (0 > (transnum = getnexttrans(name)))
    {
      free(a->ptr);
      w_error(c, ERR_MSG_SEND_TRANS, errno);
      return;
    }

  trans = open(name, O_CREAT | O_WRONLY, 0600);
  if (trans < 0)
    {
      free(a->ptr);
      w_error(c, ERR_MSG_SEND_OPEN, errno);
      return;
    }

  ptr = a->ptr + 4; /* skip over request; remember to free it when done! */

  /* authenticated username */
  r = write(trans, buffer, (l = sprintf(buffer, "%s\n", a->ad.pname)));
  if (r != l)
    {
      free(a->ptr);
      w_error(c, ERR_MSG_SEND_WRITE, errno);
      close(trans); /* don't care if this loses; it shouldn't */
      unlink(name);
      return;
    }

  /* authentication level; will always be 1 for now */
  r = write(trans, buffer, (l = sprintf(buffer, "%d\n", a->authenticated)));
  if (r != l)
    {
      free(a->ptr);
      w_error(c, ERR_MSG_SEND_WRITE, errno);
      close(trans);
      unlink(name);
      return;
    }

  /* usernames with seen/notseen flag */
  while ((int)(e = index(ptr, '\n')) != 0) /* what if no 0 ever...  bug */
    {
      r = write(trans, buffer, (l = sprintf(buffer, "%.*s 0\n", e-ptr, ptr)));
      if (r != l)
	{
	  free(a->ptr);
	  w_error(c, ERR_MSG_SEND_WRITE, errno);
	  close(trans);
	  unlink(name);
	  return;
	}
      ptr = e + 1;
    }

  /* advance to start of application data; it can be anything at all */
  while (*ptr != '\0')
    ptr++;
  ptr++;

  /* output delimiter */
  r = write(trans, buffer, (l = sprintf(buffer, "-\n")));
  if (r != l)
    {
      free(a->ptr);
      w_error(c, ERR_MSG_SEND_WRITE, errno);
      close(trans);
      unlink(name);
      return;
    }

  r = write(trans, ptr, l = (a->len - (ptr - a->ptr)));
  if (r != l)
    {
      free(a->ptr);
      w_error(c, ERR_MSG_SEND_WRITE, errno);
      close(trans);
      unlink(name);
      return;
    }

  close(trans); /* close shouldn't lose. */

  /* and we're done! */
  free(a->ptr);
  w_error(c, ERR_MSG_SEND_SUCCESS, transnum);

  ferr = fork();
  if (ferr == -1)
    fprintf(stderr, "(%d) fork for distribution of message %d failed: %d\n",
	    c->context, transnum, errno);
  else
    if (ferr == 0)
      {
	sprintf(buffer, "%d", transnum);
	if (-1 == execl(DISTFULLNAME, DISTNAME, name, buffer, 0))
	  fprintf(stderr,
		  "(%d) execl for distribution of message %d failed: %d\n",
		  c->context, transnum, errno);
      }
}

/*
 * First part of REQ_AUTHENTICATE
 * Attempt to authenticate the client
 */
void getauth(a, data)
context *a;
char *data;
{
  int k_err;
  char *instance[50];

  sprintf(instance, "*");

  if (RD_AP_OK != (k_err = krb_rd_req((KTEXT)(data),
				      "rcmd",
				      instance,
				      0, /* a->c->peer.sin_addr.s_addr, */
				      &a->ad,
				      "")))
    fprintf(stdout, "(%d) Authentication request failed:\n   %s\n",
	    a->c->context, krb_err_txt[k_err]);
  else
    {
      fprintf(stdout, "(%d) Principal %s authenticated\n",
	      a->c->context, a->ad.pname);
      a->authenticated = 1;
    }

  a->sub_err = k_err;

  /* respond with mutual authentication or error */
  a->c->s_write = sendauth;
  ungag(a->c);
}

/*
 * Second part of REQ_AUTHENTICATE
 * If client was successfully authenticated, attempt to mutually
 * authenticate.
 */
void sendauth(c)
connection *c;
{
  context *a;
  int ret;

  a = contexts[c->context];

  a->c->w_buffer = (char *)&a->srv_err;

  if (a->sub_err == RD_AP_OK)
    {
      des_key_sched(a->ad.session, a->sched);
      ret = htonl(a->ad.checksum + 1);

      if (-1 == (a->c->w_len = krb_mk_priv(&ret,             /* in */
					   a->miscbuff,      /* out */
					   sizeof(ret),      /* in_length */
					   a->sched,         /* schedule */
					   a->ad.session,    /* key */
					   &a->c->us,        /* sender */
					   &a->c->peer)))    /* receiver */
	{
	  fprintf(stdout,
	      "(%d) mk_priv returned error; mutual authentication failed.\n");
	  a->sub_err = htonl(0);
	  a->srv_err = htonl(ERR_AUTHENTICATE_MUTUAL);
	  a->c->w_len = 2 * sizeof(int);
	}
      else
	{
	  a->sub_err = htonl(0);
	  a->srv_err = htonl(ERR_AUTHENTICATE_SUCCESS);
	  a->c->w_len += 2 * sizeof(int);
	}
    }
  else
    {
      a->sub_err = htonl(a->sub_err);
      a->srv_err = htonl(ERR_AUTHENTICATE_FAILURE);
      a->c->w_len = 2 * sizeof(int);
    }

  gag(a->c);
  a->c->s_write = NULL;
}

void echo(c)
connection *c;
{
  c->s_empty = callfree;

  gag(c);
  c->s_write = NULL;
}

/*
 * Somebody is asking us to do something...
 * Here we do it, and set up for procreqw to reply with the
 * requested information/error code
 */
void procreq(c)
connection *c;
{
  context *a;
  int req;

  a = contexts[c->context];

  req = ntohl(*(int *)(a->c->r_buffer));
  fprintf(stderr, "(%d) requests %d\n", c->context, req);

  switch(req)
    {
    case REQ_ECHO:
      c->w_buffer = c->r_buffer;
      c->w_len = c->r_len;
      a->c->s_write = echo;
      ungag(a->c);
      break;

    case REQ_GROUPS_ALL:
      a->c->s_write = reqgroupsall;
      ungag(a->c);

      free(a->c->r_buffer);
      a->c->r_buffer = NULL;
      break;

    case REQ_MSG_SEND:
      a->ptr = a->c->r_buffer;
      a->len = a->c->r_len;
      a->c->s_write = msgsend;
      ungag(a->c);

      break;

    case REQ_MSG_RECV:
      a->c->s_write = msgrecv;
      ungag(a->c);

      free(a->c->r_buffer);
      a->c->r_buffer = NULL;
      break;

    case REQ_AUTHENTICATE:
      getauth(a, a->c->r_buffer + sizeof(req));

      free(a->c->r_buffer);
      a->c->r_buffer = NULL;
      break;

    default:
      fprintf(stderr, "(%d) Unknown request %d ignored\n", c->context, req);

      free(a->c->r_buffer);
      a->c->r_buffer = NULL;
      break;
    }
}

void deadsock(c)
connection *c;
{
  fprintf(stdout, "(%d) Connection closed.\n", c->context);
}

main()
{
  connection *listener;

  i_initialize();
  listener = i_create();

  listener->s_new = newclient;
  listener->s_read = procreq;
  listener->s_dead = deadsock;

  if (I_SUCCESS != i_listen(listener))
    {
      fprintf(stdout, "listen: %s\n", errtext);
      exit(0);
    }

  while (1)
    {
      i_process(60);
    }
}
